/* @(#) memory.c: mem handling stuff
 *
 * 2/11/99 JC
 *	- from im_open.c and callback.c
 *	- malloc tracking stuff added
 * 11/3/01 JC
 * 	- im_strncpy() added
 * 20/4/01 JC
 * 	- im_(v)snprintf() added
 */

/*

    This file is part of VIPS.
    
    VIPS is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

/*

    These files are distributed with VIPS - http://www.vips.ecs.soton.ac.uk

 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /*HAVE_CONFIG_H*/
#include <vips/intl.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>

#include <vips/vips.h>

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif /*WITH_DMALLOC*/

/* Define for simple malloc tracking ... better to use dmalloc if you can.
#define DEBUGM
 */

/* Track total alloc/total free here for debugging.
 */
#ifdef DEBUGM
static unsigned int total_mem_alloc = 0;
static unsigned int total_allocs = 0;
static unsigned int high_water_mark = 0;
static GMutex *malloc_mutex = NULL;
#endif /*DEBUGM*/

/* VIPS free function. Try to put all vips free() through this.
 */
int
im_free( void *s )
{
#ifdef DEBUGM
	int size;

	s = (void *) ((char*)s - 16);
	size = *((unsigned int*)s);
	g_mutex_lock( malloc_mutex );
	total_mem_alloc -= size;
	total_allocs -= 1;
	printf( "im_free: %d, %d allocs, total %.3gM, high water %.3gM\n", 
		size,
		total_allocs,
		total_mem_alloc / (1024.0 * 1024.0), 
		high_water_mark / (1024.0 * 1024.0) );

	if( total_allocs < 0 ) {
		printf( "woah! too many im_free()s!\n" );
		abort();
	}
	g_mutex_unlock( malloc_mutex );
#endif /*DEBUGM*/

	free( s );

	return( 0 );
}

/* Malloc local to a descriptor. Try to put all vips malloc through this. Not
 * thread-safe if im != NULL.
 */
void *
im_malloc( IMAGE *im, int size )
{
        void *buf;

#ifdef DEBUGM
	/* Assume the first im_malloc() is single-threaded.
	 */
	if( !malloc_mutex )
		malloc_mutex = g_mutex_new();
#endif /*DEBUGM*/

	if( size <= 0 ) {
#ifdef DEBUG
		error_exit( "im_malloc: size <= 0" );
#endif /*DEBUG*/
		im_warning( "im_malloc: size <= 0, expect a crash soon!" );
	}

#ifdef DEBUGM
	/* If debugging mallocs, need an extra sizeof(uint) bytes to track 
	 * size of this block. Ask for an extra 16 to make sure we don't break
	 * alignment rules.
	 */
	size += 16;
#endif /*DEBUGM*/

        if( !(buf = malloc( size )) ) {
#ifdef DEBUG
		/* Make sure we catch this ... 
		 */
		error_exit( "im_malloc: out of memory" );
#endif /*DEBUG*/
		im_warning( "malloc failed, very low on memory" );
		im_errormsg( "im_malloc: out of memory --- "
			"size == %dMB", (int) (size / (1024.0*1024.0)) );
                return( NULL );
	}

#ifdef DEBUGM
	/* Record number alloced.
	 */
	*((unsigned int*)buf) = size;
	buf = (void *) ((char*)buf + 16);
	g_mutex_lock( malloc_mutex );
	total_mem_alloc += size;
	if( total_mem_alloc > high_water_mark ) 
		high_water_mark = total_mem_alloc;
	total_allocs += 1;
	printf( "im_malloc: %d, %d allocs, total %.3gM, high water %.3gM\n", 
		size, 
		total_allocs,
		total_mem_alloc / (1024.0 * 1024.0),
		high_water_mark / (1024.0 * 1024.0) );
	g_mutex_unlock( malloc_mutex );

	/* Handy to breakpoint on this printf() for catching large mallocs().
	 */
	if( size > 1000000 ) 
		printf( "woah! big!\n" );
#endif /*DEBUGM*/
 
        if( im && im_add_close_callback( im, 
		(im_callback_fn) im_free, buf, NULL ) ) {
                im_free( buf );
                return( NULL );
        }
 
        return( buf );
}
